#ifndef MSP_GL_VERTEXARRAY_H_
#define MSP_GL_VERTEXARRAY_H_

#include <climits>
#include <vector>
#include <msp/core/refptr.h>
#include <msp/datafile/loader.h>
#include "bindable.h"
#include "bufferable.h"
#include "datatype.h"
#include "primitivetype.h"
#include "vertexarraybuilder.h"
#include "vertexformat.h"

namespace Msp {
namespace GL {

class Buffer;

/**
Stores vertex data.  Both legacy and generic attributes are supported.  Mixing
the two is possible but discouraged, as driver-specific issues may arise.

The array's contents can be modified with the append and modify methods.  To
obtain the location of an individual component within the vertex, use
VertexFormat::offset.

A higher-level interface for filling in vertex data is available in the
VertexArrayBuilder class.
*/
class VertexArray: public Bindable<VertexArray>, public Bufferable
{
public:
	class Loader: public DataFile::Loader, public VertexArrayBuilder
	{
	public:
		Loader(VertexArray &);
	};

private:
	struct Array
	{
		unsigned char component;
		unsigned char offset;

		Array();
	};

	VertexFormat format;
	std::vector<float> data;
	unsigned stride;
	std::vector<Array> arrays;
	bool legacy;

	static bool legacy_used;

	VertexArray(const VertexArray &);
	VertexArray &operator=(const VertexArray &);
public:
	VertexArray(const VertexFormat &);
	~VertexArray();

	/// Resets the VertexArray to a different format.  All data is cleared.
	void reset(const VertexFormat &);

	const VertexFormat &get_format() const { return format; }
private:
	static unsigned get_array_slot(unsigned char);

public:
	/// Deprecated alias for use_buffer
	void use_vertex_buffer(Buffer *b) { use_buffer(b); }

	/// Clears all vertices from the array.
	void clear();

	/// Reserve space for vertices.
	void reserve(unsigned);

	/// Append a new vertex at the end of the array and return its location.
	float *append();

	/// Returns the location of a vertex for modification.
	float *modify(unsigned);
private:
	virtual unsigned get_data_size() const;
	virtual const void *get_data_pointer() const { return &data[0]; }

public:
	unsigned size() const { return data.size()/stride; }
	const std::vector<float> &get_data() const { return data; }
	const float *operator[](unsigned i) const { return &data[0]+i*stride; }

	/** Equivalent to apply(true).  For compatibility with the Bindable
	interface. */
	void bind() const { apply(); }

	/** Applies component arrays to the GL.  If legacy is true, they are applied
	as is.  If legacy is false, any legacy attributes are converted to generic
	attributes. */
	void apply(bool legacy = true) const;

private:
	static void apply_arrays(const std::vector<Array> *, const std::vector<Array> *, const float *, unsigned, bool);
public:
	static void unbind();
};

} // namespace GL
} // namespace Msp

#endif
